''
'' This program controls the X and Y servos of a gimballed rocket motor
'' The servo pins are defined by the constants _xServoPin and _yServoPin
'' This program also sends the rotation matrix to the PropViz program
'' using a serial port.  The transmit serial pin is defined by _xmtSerialPin
''
'' Written by Dave Hein.  Based on Jason Dorie's DCM QuadRotor program
''               

{
  TODO:
        Use integrated gyro value for P, current gyro measurement for D
        Use accelerometer to bias integrated gyro result
                Store last 128 results in HUB RAM, keep a running sum, SHR 7 for current value

  DONE:
        Temperature drift compensation
        Convert ITG-3200 code to PASM
        Add ADXL345 readings to ITG code
        Integration of gyro values
}


CON
  _clkmode = xtal1 + pll8x
  _xinfreq = 10_000_000
  _xServoPin = 12
  _yServoPin = 13
  _xmtSerialPin = 18


OBJ
  Gyro  : "9DOF-pasm.spin"                              ' 1 cog
  'Flt   : "Float32Full-Limited_Routines-1_Cog"          ' 1 cog                        
  Dbg   : "FullDuplexSerial.spin"                       ' 1 cog
  Xmt   : "FullDuplexSerial.spin"                       ' 1 cog

  'Main loop                                            ' 1 cog

  DCM   : "PropDCM28 xyz"                               ' 0 cogs
  PWM   : "PWM2C_SEREngine.spin"                        ' 1 cog                          

VAR
  long counter
  long LastTime, CurTime

  long dGx, dGy, dGz                                    ' Working variables used by the DCM code
  long dAx, dAy, dAz

  long XAngle, YAngle, ZAngle

  long stack[100]



PUB Main | v, x , y
 
  Xmt.Start( 17, _xmtSerialPin, 0, 115200 )
  Dbg.Start( 31, 30, 0, 115200 )
  Gyro.Start( 20, 21 )
  'Flt.Start
  cognew(TransmitThread, @stack)
  pwm.SEREngineStart(_xServoPin, _yServoPin, 50)
    
  RunDCM  'Can be put into its own cog if desired




PUB RunDCM | tx, tz, loopTime

  DCM.Init

  'Start by gathering an average accelerometer value to use as "down"
  'Assumes the board is flat / level on startup
  dAx := 0
  dAy := 0
  dAz := 0
  loopTime := cnt

  repeat counter from 1 to 256
    dAx += Gyro.GetAX
    dAy += Gyro.GetAY
    dAz += Gyro.GetAZ

    waitcnt( constant(80_000_000 / 200) + loopTime )
    loopTime += constant(80_000_000 / 200)

  dAx := dAx << (DCM#FixedBits - (8 + 8))               '8 bits of accel value, 8 bits of loop accumulation
  dAy := dAy << (DCM#FixedBits - (8 + 8))
  dAz := dAz << (DCM#FixedBits - (8 + 8))
  

  'BiasCorrect SHOULD NOT include 1G gravity
  DCM.SetBiasCorrect( dAx, dAy, dAz - DCM#Fixed_One )

  loopTime := cnt

  repeat
    dGx := Gyro.GetRX
    dGy := Gyro.GetRY
    dGz := Gyro.GetRZ
    dAx := Gyro.GetAX
    dAy := Gyro.GetAY
    dAz := Gyro.GetAZ
    DCM.SetValues( dGx, dGy, dGz, dAx, dAy, dAz )
    DCM.Update
    ServoUpdate
    waitcnt( constant(80_000_000 / 200) + loopTime )
    loopTime += constant(80_000_000 / 200)

PUB TransmitThread
  repeat
    TransmitMatrix
    PrintMatrix
    waitcnt(clkfreq/20+cnt)

PUB TransmitMatrix | i, val
    repeat i from 0 to 8
      val := DCM.GetMatrixValue(i)
      val := (val + (1 << 20)) ~> 21
      if val > 127
        val := 127
      if val < -128
        val := -128
      xmt.tx(val)
    xmt.tx(13)

PUB PrintMatrix | i, j, val, sign, divisor, digit
    dbg.tx(0)
    repeat i from 0 to 8
      val := DCM.GetMatrixValue(i)
      if val => 0
        sign := 0
      else
        sign := 1
        val := -val
      val := (val + 1342177) / 2684355
      ifnot val
        sign := 0
      if sign
        dbg.tx("-")
      else
        dbg.tx(" ")
      divisor := 100
      repeat j from 1 to 3
        digit := val/divisor
        val //= divisor
        divisor /= 10
        dbg.tx(digit+"0")
        if j == 1
          dbg.tx(".")
      if i // 3 == 2
        dbg.tx(13)
      else
        dbg.tx(" ")

PUB ServoUpdate | xpulse, ypulse
  xpulse := (DCM.GetMatrixValue(6) ~> (28 - 9)) + 1500
  if xpulse < 1000
    xpulse := 1000
  if xpulse > 2000
    xpulse := 2000
  ypulse := (DCM.GetMatrixValue(7) ~> (28 - 9)) + 1500
  if ypulse < 1000
    ypulse := 1000
  if ypulse > 2000
    ypulse := 2000
  PWM.leftPulseLength(xpulse)
  PWM.rightPulseLength(ypulse)
    
            